--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity CTRLREG is
	port (
		ADD			: in  std_logic_vector(9 downto 0);		-- from CPU
		DATAI		: in  std_logic_vector(7 downto 0);		-- from CPU
		WRN			: in  std_logic;						-- from CPU
		RDN			: in  std_logic;						-- from CPU
		LCDMODE		: in  std_logic;
		MK2MODE		: in  std_logic;
		FUNCKEY		: in  std_logic_vector(23 downto 0);
		CLK50M		: in  std_logic;
		RSTN		: in  std_logic;
		DATAO		: out std_logic_vector(7 downto 0);		-- to CPU
		LCDCS		: out std_logic;						-- to LCD controller
		LCDSDI		: out std_logic;						-- to LCD controller
		LCDSCL		: out std_logic;						-- to LCD controller
		LCDDONE		: out std_logic;
		CTRLSEL		: out std_logic;
		CTRLRSTN	: out std_logic;
		MEM16K		: out std_logic;
		SC4COLORON	: out std_logic;
		SC4COLORMD	: out std_logic_vector(3 downto 0);
		TAPEACC		: out std_logic;
		TAPEACCH	: out std_logic
	);
end CTRLREG;

architecture RTL of CTRLREG is

	signal reg_lcdcs		: std_logic;					-- 0x8401(0)
	signal reg_lcdsdi		: std_logic;					-- 0x8401(1)
	signal reg_lcdscl		: std_logic;					-- 0x8401(2)
	signal reg_lcddone		: std_logic;					-- 0x8402(0)

	signal reg_ctrlselcpu	: std_logic;					-- 0x8410(0)
	signal reg_ctrlsel		: std_logic;					-- 0x8410(1)
	signal reg_ctrlrstcpu	: std_logic;					-- 0x8411(0)

	signal reg_mem16k		: std_logic;					-- 0x8412(0)
	signal reg_sc4coloron	: std_logic;					-- 0x8413(4)
	signal reg_sc4colormd	: std_logic_vector(3 downto 0);	-- 0x8413(3:0)
	signal reg_tapeacc		: std_logic;					-- 0x8414(0)
	signal reg_tapeacch		: std_logic;					-- 0x8414(1)

	signal ctrlsel_i		: std_logic;
	signal ctrlrstn_i		: std_logic;
	signal ctrlrst_f1		: std_logic;
	signal ctrlrstcnt		: std_logic_vector(11 downto 0);

	signal datao_i			: std_logic_vector(7 downto 0);

	signal rdn_f1			: std_logic;
	signal rdn_f2			: std_logic;
	signal rdn_f3			: std_logic;
	signal wrn_f1			: std_logic;
	signal wrn_f2			: std_logic;
	signal wrn_f3			: std_logic;
	signal add_f1			: std_logic_vector(9 downto 0);
	signal add_f2			: std_logic_vector(9 downto 0);
	signal datai_f1			: std_logic_vector(7 downto 0);
	signal datai_f2			: std_logic_vector(7 downto 0);
	signal rdp				: std_logic;
	signal wrp				: std_logic;

	signal funckey_f1		: std_logic_vector(23 downto 0);
	signal funckey_f2		: std_logic_vector(23 downto 0);
	signal funckey_f3		: std_logic_vector(23 downto 0);

begin

-- CPU I/F
-- data latch
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			rdn_f1   <= '1';
			rdn_f2   <= '1';
			rdn_f3   <= '1';
			wrn_f1   <= '1';
			wrn_f2   <= '1';
			wrn_f3   <= '1';
			add_f1   <= (others => '0');
			add_f2   <= (others => '0');
			datai_f1 <= (others => '0');
			datai_f2 <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			rdn_f1   <= RDN;
			rdn_f2   <= rdn_f1;
			rdn_f3   <= rdn_f2;
			wrn_f1   <= WRN;
			wrn_f2   <= wrn_f1;
			wrn_f3   <= wrn_f2;
			add_f1   <= ADD;
			add_f2   <= add_f1;
			datai_f1 <= DATAI;
			datai_f2 <= datai_f1;
		end if;
	end process;

	rdp <= '1' when (rdn_f2 = '0' and rdn_f3 = '1') else '0';
	wrp <= '1' when (wrn_f2 = '0' and wrn_f3 = '1') else '0';

-- register
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			reg_lcdcs      <= '1';
			reg_lcdsdi     <= '1';
			reg_lcdscl     <= '1';
			reg_lcddone    <= '0';
			reg_ctrlselcpu <= '0';
			reg_ctrlsel    <= '0';
			reg_ctrlrstcpu <= '0';
			reg_mem16k     <= '0';
			reg_sc4coloron <= '1';
			reg_sc4colormd <= "0000";
			reg_tapeacc    <= '1';
			reg_tapeacch   <= '1';
		elsif (CLK50M'event and CLK50M = '1') then
			if (wrp = '1') then
				case add_f2 is
					when "0000000001"=>	reg_lcdcs      <= datai_f2(0);
										reg_lcdsdi     <= datai_f2(1);
										reg_lcdscl     <= datai_f2(2);
					when "0000000010"=>	reg_lcddone    <= datai_f2(0);
					when "0000010000"=>	reg_ctrlselcpu <= datai_f2(0);
										reg_ctrlsel    <= datai_f2(1);
					when "0000010001"=>	reg_ctrlrstcpu <= datai_f2(0);
					when "0000010010"=> reg_mem16k     <= datai_f2(0);
					when "0000010011"=> reg_sc4coloron <= datai_f2(4);
										reg_sc4colormd <= datai_f2(3 downto 0);
					when "0000010100"=> reg_tapeacc    <= datai_f2(0);
										reg_tapeacch   <= datai_f2(1);
					when others      => null;
				end case;
			end if;
		end if;
	end process;

-- register read
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			datao_i <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			if (rdp = '1') then
				case add_f2 is
					when "0000000000"=> datao_i <= "0000000" & LCDMODE;
					when "0000000001"=> datao_i <= "00000" & reg_lcdscl & reg_lcdsdi & reg_lcdcs;
					when "0000000010"=> datao_i <= "0000000" & reg_lcddone;
					when "0000010000"=> datao_i <= "00000" & ctrlsel_i & reg_ctrlsel & reg_ctrlselcpu;
					when "0000010001"=> datao_i <= "0000000" & reg_ctrlrstcpu;
					when "0000010010"=> datao_i <= "0000000" & reg_mem16k;
					when "0000010011"=> datao_i <= "000" & reg_sc4coloron & reg_sc4colormd;
					when "0000010100"=> datao_i <= "000000" & reg_tapeacch & reg_tapeacc;
					when "0000010101"=> datao_i <= "0000000" & MK2MODE;
					when "0000100000"=> datao_i <= funckey_f2( 7 downto  0);
					when "0000100001"=> datao_i <= funckey_f2(15 downto  8);
					when "0000100010"=> datao_i <= funckey_f2(23 downto 16);
					when others      => datao_i <= X"FF";
				end case;
			end if;
		end if;
	end process;


	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			funckey_f1 <= (others => '0');
			funckey_f2 <= (others => '0');
			funckey_f3 <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			funckey_f1 <= FUNCKEY;
			funckey_f2 <= funckey_f1;
			funckey_f3 <= funckey_f2;
		end if;
	end process;


	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			ctrlsel_i <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_ctrlselcpu = '1') then
				ctrlsel_i <= reg_ctrlsel;
			elsif (funckey_f2(8) = '1' and funckey_f3(8) = '0') then
				ctrlsel_i <= not ctrlsel_i;
			end if;
		end if;
	end process;


	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			ctrlrst_f1 <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_ctrlrstcpu = '1') then
				ctrlrst_f1 <= '1';
			elsif (funckey_f2(11) = '1' and funckey_f3(11) = '0') then
				ctrlrst_f1 <= '1';
			else
				ctrlrst_f1 <= '0';
			end if;
		end if;
	end process;

	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			ctrlrstcnt <= X"FFF";
			ctrlrstn_i <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (ctrlrst_f1 = '1') then
				ctrlrstcnt <= X"000";
				ctrlrstn_i <= '0';
			elsif (ctrlrstcnt = X"FFF") then
				ctrlrstn_i <= '1';
			else
				ctrlrstcnt <= ctrlrstcnt + 1;
				ctrlrstn_i <= '0';
			end if;
		end if;
	end process;



	DATAO      <= datao_i;

	LCDCS      <= reg_lcdcs;
	LCDSDI     <= reg_lcdsdi;
	LCDSCL     <= reg_lcdscl;
	LCDDONE    <= reg_lcddone;

	CTRLSEL    <= ctrlsel_i;
	CTRLRSTN   <= ctrlrstn_i;

	MEM16K     <= reg_mem16k;
	SC4COLORON <= reg_sc4coloron;
	SC4COLORMD <= reg_sc4colormd;
	TAPEACC    <= reg_tapeacc;
	TAPEACCH   <= reg_tapeacch;

end RTL;
